#ifndef BASE_ALGORITHM_H
#define BASE_ALGORITHM_H

#include "functional.h"
#include <algorithm>
#include <array>
#include <cstddef>
#include <functional>
#include <iterator>
#include <type_traits>

namespace base {
   template<typename II, typename F>
   inline void for_each(II ai, II af, const F& f)
   {
      std::for_each(ai, af, f);
   }

   template<typename II1, typename II2, typename F>
   inline void for_each(II1 ai, II1 af, II2 bi, const F& f)
   {
      while (ai != af) {
         f(*ai++, *bi++);
      }
   }


   template<typename RI, typename T, typename PRED>
   inline RI binary_find(RI i, RI f, const T& v, const PRED& p)
   {
      auto res = std::lower_bound(i, f, v, p);
      return (res != f && !p(v, *res) ? res : f);
   }

   template<typename RI, typename T>
   inline RI binary_find(RI i, RI f, const T& v)
   {
      return binary_find(i, f, v, std::less<>( ));
   }


   template<typename RI1, typename RI2, typename OI, typename PRED>
   inline OI binary_multisearch(RI1 ai, RI1 af, RI2 bi, RI2 bf, OI wi, const PRED& p)
   {
      if (bi == bf) {
         return wi;
      }

      if (ai == af) {
         return std::fill_n(wi, bf - bi, af);
      }

      auto bj = bi + (bf - bi) / 2;
      auto aj = std::lower_bound(ai, af, *bj, p);

      wi = binary_multisearch(ai, aj, bi, bj, wi, p);
      *wi++ = aj;
      wi = binary_multisearch(aj, af, bj + 1, bf, wi, p);

      return wi;
   }

   template<typename RI1, typename RI2, typename OI>
   inline OI binary_multisearch(RI1 ai, RI1 af, RI2 bi, RI2 bf, OI wi)
   {
      return binary_multisearch(ai, af, bi, bf, wi, std::less<>( ));
   }


   template<typename T, typename PRED>
   inline int compare_3way(const T& a, const T& b, const PRED& p)
   {
      return p(b, a) - p(a, b);
   }

   template<typename T>
   inline int compare_3way(const T& a, const T& b)
   {
      return compare_3way(a, b, std::less<T>( ));
   }


   template<typename FI1, typename FI2, typename PRED>
   inline bool memory_compare(FI1 ai, FI1 af, FI2 bi, const PRED& p)
   {
      while (ai != af) {
         if (p(*ai, *bi)) {
            return true;
         }

         if (p(*bi, *ai)) {
            return false;
         }

         ++ai;
         ++bi;
      }

      return false;
   }

   template<typename FI1, typename FI2>
   inline bool memory_compare(FI1 ai, FI1 af, FI2 bi)
   {
      return memory_compare(ai, af, bi, std::less<typename std::common_type<typename std::iterator_traits<FI1>::value_type, typename std::iterator_traits<FI2>::value_type>::type>( ));
   }

   template<typename FI1, typename FI2, typename PRED>
   inline int memory_compare_3way(FI1 ai, FI1 af, FI2 bi, const PRED& p)
   {
      while (ai != af) {
         auto res = compare_3way(*ai, *bi, p);

         if (res != 0) {
            return res;
         }

         ++ai;
         ++bi;
      }

      return 0;
   }

   template<typename FI1, typename FI2>
   inline int memory_compare_3way(FI1 ai, FI1 af, FI2 bi)
   {
      return memory_compare_3way(ai, af, bi, std::less<typename std::common_type<typename std::iterator_traits<FI1>::value_type, typename std::iterator_traits<FI2>::value_type>::type>( ));
   }

   template<typename FI1, typename FI2, typename PRED>
   inline int lexicographical_compare_3way(FI1 ai, FI1 af, FI2 bi, FI2 bf, const PRED& p)
   {
      while (ai != af && bi != bf) {
         auto res = compare_3way(*ai, *bi, p);

         if (res != 0) {
            return res;
         }

         ++ai;
         ++bi;
      }

      return (ai != af) - (bi != bf);
   }

   template<typename FI1, typename FI2>
   inline int lexicographical_compare_3way(FI1 ai, FI1 af, FI2 bi, FI2 bf)
   {
      return lexicographical_compare_3way(ai, af, bi, bf, std::less<typename std::common_type<typename std::iterator_traits<FI1>::value_type, typename std::iterator_traits<FI2>::value_type>::type>( ));
   }


   namespace impl {
      template<typename FI, typename PRED>
      inline FI find_last_if(FI ai, FI af, const PRED& p, std::forward_iterator_tag)
      {
         auto res = af;

         for (;;) {
            ai = std::find_if(ai, af, p);

            if (ai == af) {
               return res;
            }

            res = ai++;
         }
      }

      template<typename BI, typename PRED>
      inline BI find_last_if(BI ai, BI af, const PRED& p, std::bidirectional_iterator_tag)
      {
         auto ini = std::reverse_iterator<BI>(af);
         auto fin = std::reverse_iterator<BI>(ai);
         auto res = std::find_if(ini, fin, p);

         return (res == fin ? af : std::prev(res.base( )));
      }
   }

   template<typename FI, typename PRED>
   inline FI find_last_if(FI ai, FI af, const PRED& p)
   {
      return impl::find_last_if(ai, af, p, typename std::iterator_traits<FI>::iterator_category( ));
   }

   template<typename FI, typename T>
   inline FI find_last(FI ai, FI af, const T& v)
   {
      return find_last_if(ai, af, std::bind(std::equal_to<>( ), std::placeholders::_1, std::cref(v)));
   }


   namespace impl {
      template<typename FI, typename BUS, typename PRED>
      inline FI find_last_of(FI ai, FI af, BUS bi, BUS bf, const PRED& p, std::forward_iterator_tag)
      {
         auto res = af;

         for (;;) {
            ai = std::find_first_of(ai, af, bi, bf, p);

            if (ai == af) {
               return res;
            }

            res = ai++;
         }
      }

      template<typename BI, typename BUS, typename PRED>
      inline BI find_last_of(BI ai, BI af, BUS bi, BUS bf, const PRED& p, std::bidirectional_iterator_tag)
      {
         auto ini = std::reverse_iterator<BI>(af);
         auto fin = std::reverse_iterator<BI>(ai);
         auto res = std::find_first_of(ini, fin, bi, bf, p);

         return (res == fin ? af : std::prev(res.base( )));
      }
   }

   template<typename FI, typename BUS, typename PRED>
   inline FI find_last_of(FI ai, FI af, BUS bi, BUS bf, const PRED& p)
   {
      return impl::find_last_of(ai, af, bi, bf, p, typename std::iterator_traits<FI>::iterator_category( ));
   }

   template<typename FI, typename BUS>
   inline FI find_last_of(FI ai, FI af, BUS bi, BUS bf)
   {
      return find_last_of(ai, af, bi, bf, std::equal_to<typename std::iterator_traits<FI>::value_type>( ));
   }


   template<typename II, typename OI, typename PRED>
   inline std::pair<II, OI> copy_while(II ai, II af, OI oi, const PRED& p)
   {
      while (ai != af && p(*ai)) {
         *oi++ = *ai++;
      }

      return { ai, oi };
   }

   template<typename II, typename OI, typename PRED>
   inline std::pair<II, OI> copy_until(II ai, II af, OI oi, const PRED& p)
   {
      return copy_while(ai, af, oi, base::not_fn(p));
   }


   namespace impl {
      template<typename FI, typename C>
      inline typename std::enable_if<C::partition_size( ) >= 4, std::array<FI, C::partition_size( ) - 1>>::type multiway_partition(FI ai, FI af, const C& cl, std::forward_iterator_tag)
      {
         using size_type = decltype(C::partition_size( ));

         std::array<FI, C::partition_size( )> ini;
         std::array<FI, C::partition_size( ) - 1> fin;
         std::fill(ini.begin( ), ini.end( ), ai);

         for (auto i = ai; i != af; ++i) {
            ++ini[cl(*i)];
         }

         auto actual = ai;

         for (size_type i = 0; i < C::partition_size( ) - 1; ++i) {
            auto cuantos = ini[i] - ai;

            ini[i] = actual;
            actual += cuantos;
            fin[i] = actual;
         }

         ini.back( ) = actual;

         for (size_type i = 0; i < C::partition_size( ) - 1; ++i) {
            auto ji = ini[i];
            auto jf = fin[i];

            while (ji < jf) {
               auto offset = cl(*ji);

               if (offset == i) {
                  ++ji;
               }
               else {
                  std::swap(*ji, *ini[offset]++);
               }
            }
         }

         return fin;
      }

      template<typename FI, typename C>
      inline typename std::enable_if<C::partition_size( ) == 3, std::array<FI, C::partition_size( ) - 1>>::type multiway_partition(FI ai, FI af, const C& cl, std::forward_iterator_tag)
      {
         FI f1 = ai;
         FI f2 = ai;

         while (ai != af) {
            auto res = cl(*ai);

            if (res == 0) {
               std::swap(*ai, *f2);
               std::swap(*f1++, *f2++);
            }
            else if (res == 1) {
               std::swap(*ai, *f2++);
            }

            ++ai;
         }

         return { f1, f2 };
      }

      template<typename BI, typename C>
      inline typename std::enable_if<C::partition_size( ) == 3, std::array<BI, C::partition_size( ) - 1>>::type multiway_partition(BI ai, BI af, const C& cl, std::bidirectional_iterator_tag)
      {
         BI f1 = ai;
         BI f2 = af;

         while (ai != f2) {
            auto res = cl(*ai);

            if (res == 2) {
               std::swap(*ai, *--f2);
            }
            else {
               if (res == 0) {
                  std::swap(*ai, *f1++);
               }

               ++ai;
            }
         }

         return { f1, f2 };
      }

      template<typename FI, typename C, typename IT>
      inline typename std::enable_if<C::partition_size( ) == 2, std::array<FI, C::partition_size( ) - 1>>::type multiway_partition(FI ai, FI af, const C& cl, IT tag)
      {
         return { std::partition(ai, af, std::negate<C>(cl)) };
      }

      template<typename FI, typename C, typename IT>
      inline typename std::enable_if<C::partition_size( ) == 1, std::array<FI, C::partition_size( ) - 1>>::type multiway_partition(FI ai, FI af, const C& cl, IT tag)
      {
         return { };
      }
   }

   template<typename FI, typename C>
   inline std::array<FI, C::partition_size( ) - 1> multiway_partition(FI ai, FI af, const C& cl)
   {
      return impl::multiway_partition(ai, af, cl, typename std::iterator_traits<FI>::iterator_category( ));
   }


   template<typename FI, typename PRED>
   inline FI stable_remove_if(FI ai, FI af, const PRED& p)
   {
      return std::remove_if(ai, af, p);
   }

   template<typename FI, typename T>
   inline FI stable_remove(FI ai, FI af, const T& v)
   {
      return stable_remove_if(ai, af, std::bind(std::equal_to<T>( ), std::placeholders::_1, std::cref(v)));
   }

   namespace impl {
      template<typename BI, typename PRED>
      inline BI remove_if(BI ai, BI af, const PRED& p, std::bidirectional_iterator_tag)
      {
         for (;;) {
            auto quitar = std::find_if(ai, af, p);

            if (quitar == af) {
               return quitar;
            }

            auto siguiente = std::next(quitar);
            auto mover = find_last_if(siguiente, af, base::not_fn(p), std::bidirectional_iterator_tag( ));

            if (mover == af) {
               return quitar;
            }

            *quitar = std::move(*mover);

            ai = siguiente;
            af = mover;
         }
      }

      template<typename FI, typename PRED>
      inline FI remove_if(FI ai, FI af, const PRED& p, std::forward_iterator_tag)
      {
         return stable_remove_if(ai, af, p);
      }
   }

   template<typename FI, typename PRED>
   inline FI remove_if(FI ai, FI af, const PRED& p)
   {
      return impl::remove_if(ai, af, p, typename std::iterator_traits<FI>::iterator_category( ));
   }

   template<typename FI, typename T>
   inline FI remove(FI ai, FI af, const T& v)
   {
      return remove_if(ai, af, std::bind(std::equal_to<T>( ), std::placeholders::_1, std::cref(v)));
   }


   template<typename T, typename PRED>
   inline void sort2(T& a, T& b, const PRED& p)
   {
      if (p(b, a)) {
         std::swap(a, b);
      }
   }

   template<typename T>
   inline void sort2(T& a, T& b)
   {
      sort2(a, b, std::less<T>( ));
   }
}

#endif
