#ifndef BASE_ARRAY_H
#define BASE_ARRAY_H

#include <algorithm>
#include <array>
#include <cstddef>
#include <initializer_list>
#include <iterator>
#include <utility>

namespace base {
   template<typename T, typename... PARAMS>
   inline constexpr std::array<T, sizeof...(PARAMS)> make_array(PARAMS&&... v)
   {
       return std::array<T, sizeof...(PARAMS)>{ std::forward<PARAMS>(v)... };
   }

   namespace impl {
      template<typename T, std::size_t I, std::size_t... S>
      struct matrix {
         using type = std::array<typename matrix<T, S...>::type, I>;
      };

      template<typename T, std::size_t I>
      struct matrix<T, I> {
         using type = std::array<T, I>;
      };
   }

   template<typename T, std::size_t... I>
   using matrix = typename impl::matrix<T, I...>::type;


   template<typename T>
   class array_view {
   public:
      using value_type = T;
      using iterator = const T*;
      using const_iterator = const T*;
      using reverse_iterator = std::reverse_iterator<iterator>;
      using const_reverse_iterator = std::reverse_iterator<const_iterator>;

      array_view( )
      : mem_(nullptr), tam_(0)
      {
      }

      array_view(const_iterator ai, const_iterator af)
      : mem_(ai), tam_(af - ai)
      {
      }

      template<typename U>
      array_view(const U& v)
      : mem_(v.data( )), tam_(v.size( ))
      {
      }

      template<std::size_t N>
      array_view(const T(&arr)[N])
      : mem_(arr), tam_(N)
      {
      }

      array_view(const std::initializer_list<T>& i)
      : mem_(i.begin( )), tam_(i.size( ))
      {
      }

      const T& operator[](std::size_t i) const
      {
         return mem_[i];
      }

      bool empty( ) const
      {
         return tam_ == 0;
      }

      std::size_t size( ) const
      {
         return tam_;
      }

      const T* data( ) const
      {
         return mem_;
      }

      const_iterator begin( ) const
      {
         return mem_;
      }

      const_iterator end( ) const
      {
         return mem_ + tam_;
      }

      const_iterator cbegin( ) const
      {
         return begin( );
      }

      const_iterator cend( ) const
      {
         return end( );
      }

      const_reverse_iterator rbegin( ) const
      {
         return const_reverse_iterator(end( ));
      }

      const_reverse_iterator rend( ) const
      {
         return const_reverse_iterator(begin( ));
      }

      const_reverse_iterator crbegin( ) const
      {
         return rbegin( );
      }

      const_reverse_iterator crend( ) const
      {
         return rend( );
      }

      const value_type& front( ) const
      {
         return mem_[0];
      }

      const value_type& back( ) const
      {
         return mem_[tam_ - 1];
      }

   private:
      const T* mem_;
      std::size_t tam_;
   };

   template<typename T>
   inline bool operator<(const array_view<T>& a, const array_view<T>& b)
   {
      return std::lexicographical_compare(a.begin( ), a.end( ), b.begin( ), b.end( ));
   }

   template<typename T>
   inline bool operator==(const array_view<T>& a, const array_view<T>& b)
   {
      return (a.size( ) == b.size( ) && std::equal(a.begin( ), a.end( ), b.begin( )));
   }

   template<typename T>
   inline bool operator>(const array_view<T>& a, const array_view<T>& b)
   {
      return (b < a);
   }

   template<typename T>
   inline bool operator<=(const array_view<T>& a, const array_view<T>& b)
   {
      return !(a > b);
   }

   template<typename T>
   inline bool operator>=(const array_view<T>& a, const array_view<T>& b)
   {
      return !(a < b);
   }

   template<typename T>
   inline bool operator!=(const array_view<T>& a, const array_view<T>& b)
   {
      return !(a == b);
   }
}

#endif
