#ifndef BASE_BITS_H
#define BASE_BITS_H

#include "algorithm.h"
#include "integer.h"
#include "intrinsic.h"
#include <algorithm>
#include <array>
#include <climits>
#include <cstddef>
#include <functional>
#include <limits>
#include <type_traits>

namespace base {
   template<typename T>
   inline constexpr T bitmask(unsigned i, unsigned n)
   {
      return ((T(n != std::numeric_limits<T>::digits) << n) - 1) << i;
   }

   template<typename T>
   inline constexpr T bitmask(const T& v, unsigned i, unsigned n)
   {
      return v & bitmask<T>(i, n);
   }

   template<typename T>
   inline constexpr T bitmask_from(const T& v, unsigned i)
   {
      return v & bitmask<T>(i, std::numeric_limits<T>::digits - i);
   }

   template<typename T>
   inline constexpr T bitmask_until(const T& v, unsigned i)
   {
      return v & bitmask<T>(0, i);
   }


   template<typename T>
   inline bool get_bit(const T& v, unsigned i)
   {
      return bitmask(v, i, 1);
   }

   template<typename T>
   inline void and_bit(T& v, unsigned i, bool b)
   {
      v &= ~bitmask<T>(i, !b);
   }

   template<typename T>
   inline void or_bit(T& v, unsigned i, bool b)
   {
      v |= bitmask<T>(i, b);
   }

   template<typename T>
   inline void xor_bit(T& v, unsigned i, bool b)
   {
      v ^= bitmask<T>(i, b);
   }

   template<typename T>
   inline void reset_bit(T& v, unsigned i)
   {
      and_bit(v, i, false);
   }

   template<typename T>
   inline void set_bit(T& v, unsigned i)
   {
      or_bit(v, i, true);
   }

   template<typename T>
   inline void flip_bit(T& v, unsigned i)
   {
      xor_bit(v, i, true);
   }

   template<typename T>
   inline void write_bit(T& v, unsigned i, bool b)
   {
      reset_bit(v, i);
      or_bit(v, i, b);
   }


   template<typename T>
   inline constexpr T get_n_bits(const T& v, unsigned i, unsigned n)
   {
      return bitmask_until(v >> i, n);
   }

   template<typename T>
   inline void and_n_bits(T& v, const uint_count_t<std::numeric_limits<T>::digits>& i, const uint_count_t<std::numeric_limits<T>::digits>& n, const T& s)
   {
      v &= ~(bitmask_until(~s, n) << i);
   }

   template<typename T>
   inline void or_n_bits(T& v, const uint_count_t<std::numeric_limits<T>::digits>& i, const uint_count_t<std::numeric_limits<T>::digits>& n, const T& s)
   {
      v |= bitmask_until(s, n) << i;
   }

   template<typename T>
   inline void xor_n_bits(T& v, const uint_count_t<std::numeric_limits<T>::digits>& i, const uint_count_t<std::numeric_limits<T>::digits>& n, const T& s)
   {
      v ^= bitmask_until(s, n) << i;
   }

   template<typename T>
   inline void reset_n_bits(T& v, unsigned i, unsigned n)
   {
      v &= ~bitmask<T>(i, n);
   }

   template<typename T>
   inline void set_n_bits(T& v, unsigned i, unsigned n)
   {
      v |= bitmask<T>(i, n);
   }

   template<typename T>
   inline void flip_n_bits(T& v, unsigned i, unsigned n)
   {
      v ^= bitmask<T>(i, n);
   }

   template<typename T>
   inline void write_n_bits(T& v, const uint_count_t<std::numeric_limits<T>::digits>& i, const uint_count_t<std::numeric_limits<T>::digits>& n, const T& s)
   {
      reset_n_bits(v, i, n);
      or_n_bits(v, i, n, s);
   }


   template<typename T>
   inline void reset_bits(T& v)
   {
      v = T(0);
   }

   template<typename T>
   inline void set_bits(T& v)
   {
      v = ~T(0);
   }

   template<typename T>
   inline void flip_bits(T& v)
   {
      v = ~v;
   }


   inline constexpr unsigned popcount(unsigned char v)
   {
      return __builtin_popcount(v);
   }

   inline constexpr unsigned popcount(unsigned short v)
   {
      return __builtin_popcount(v);
   }

   inline constexpr unsigned popcount(unsigned v)
   {
      return __builtin_popcount(v);
   }

   inline constexpr unsigned popcount(unsigned long v)
   {
      return __builtin_popcountl(v);
   }

   inline constexpr unsigned popcount(unsigned long long v)
   {
      return __builtin_popcountll(v);
   }

   template<typename T>
   inline constexpr typename std::enable_if<(std::numeric_limits<T>::digits > std::numeric_limits<unsigned long long>::digits), unsigned>::type popcount(T v)
   {
      unsigned res = 0;

      while (v != 0) {
         res += popcount(static_cast<unsigned long long>(v));
         v >>= std::numeric_limits<unsigned long long>::digits;
      }

      return res;
   }


   inline void byte_swap(std::uint8_t& v)
   {
      return;
   }

   inline void byte_swap(std::uint16_t& v)
   {
      v = __builtin_bswap16(v);
   }

   inline void byte_swap(std::uint32_t& v)
   {
      v = __builtin_bswap32(v);
   }

   inline void byte_swap(std::uint64_t& v)
   {
      v = __builtin_bswap64(v);
   }

   template<typename T>
   inline void byte_swap(T& v)
   {
      auto inicio = reinterpret_cast<std::uint8_t*>(&v);
      std::reverse(inicio, inicio + sizeof(v));
   }


   inline constexpr unsigned find_first_set_bit(unsigned char v)
	{
		return (v != 0 ? __builtin_ctz(v) : std::numeric_limits<unsigned char>::digits);
	}

   inline constexpr unsigned find_first_set_bit(unsigned short v)
	{
		return (v != 0 ? __builtin_ctz(v) : std::numeric_limits<unsigned short>::digits);
	}

   inline constexpr unsigned find_first_set_bit(unsigned v)
	{
		return (v != 0 ? __builtin_ctz(v) : std::numeric_limits<unsigned>::digits);
	}

   inline constexpr unsigned find_first_set_bit(unsigned long v)
	{
		return (v != 0 ? __builtin_ctzl(v) : std::numeric_limits<unsigned long>::digits);
	}

   inline constexpr unsigned find_first_set_bit(unsigned long long v)
	{
		return (v != 0 ? __builtin_ctzll(v) : std::numeric_limits<unsigned long long>::digits);
	}

   template<typename T>
   inline constexpr typename std::enable_if<(std::numeric_limits<T>::digits > std::numeric_limits<unsigned long long>::digits), unsigned>::type find_first_set_bit(const T& v)
   {
      for (auto i = 0; i < std::numeric_limits<T>::digits; i += std::numeric_limits<unsigned long long>::digits) {
         auto temp = static_cast<unsigned long long>(v >> i);

         if (temp != 0) {
            return i + find_first_set_bit(temp);
         }
      }

      return std::numeric_limits<T>::digits;
   }

   template<typename T>
   inline unsigned find_next_set_bit(T v, unsigned i)
   {
      reset_n_bits(v, 0, i + 1);
      return find_first_set_bit(v);
   }


   inline constexpr unsigned find_last_set_bit(unsigned char v)
	{
		return (v != 0 ? std::numeric_limits<unsigned>::digits - __builtin_clz(v) - 1 : std::numeric_limits<unsigned char>::digits);
	}

   inline constexpr unsigned find_last_set_bit(unsigned short v)
	{
		return (v != 0 ? std::numeric_limits<unsigned>::digits - __builtin_clz(v) - 1 : std::numeric_limits<unsigned short>::digits);
	}

   inline constexpr unsigned find_last_set_bit(unsigned v)
	{
      return (v != 0 ? std::numeric_limits<unsigned>::digits - __builtin_clz(v) - 1 : std::numeric_limits<unsigned>::digits);
	}

   inline constexpr unsigned find_last_set_bit(unsigned long v)
	{
      return (v != 0 ? std::numeric_limits<unsigned long>::digits - __builtin_clzl(v) - 1 : std::numeric_limits<unsigned long>::digits);
	}

   inline constexpr unsigned find_last_set_bit(unsigned long long v)
	{
      return (v != 0 ? std::numeric_limits<unsigned long long>::digits - __builtin_clzll(v) - 1 : std::numeric_limits<unsigned long long>::digits);
	}

   template<typename T>
   inline constexpr typename std::enable_if<(std::numeric_limits<T>::digits > std::numeric_limits<unsigned long long>::digits), unsigned>::type find_last_set_bit(const T& v)
   {
      for (auto i = std::numeric_limits<T>::digits; i > 0; i -= std::numeric_limits<unsigned long long>::digits) {
         auto temp = static_cast<unsigned long long>(v >> (i - std::numeric_limits<unsigned long long>::digits));

         if (temp != 0) {
            return i - std::numeric_limits<unsigned long long>::digits + find_last_set_bit(temp);
         }
      }

      return std::numeric_limits<T>::digits;
   }

   template<typename T>
   inline constexpr unsigned find_previous_set_bit(T v, unsigned i)
   {
      reset_n_bits(v, i, std::numeric_limits<T>::digits - i);
      return find_last_set_bit(v);
   }


   template<typename T>
   inline constexpr unsigned find_ith_set_bit(T v, unsigned i)
   {
      auto res = 0;

      for (auto bits = std::numeric_limits<T>::digits; bits > 1; bits /= 2) {
         auto cuantos = popcount(bitmask_until(v, bits / 2));

         if (cuantos <= i) {
            i -= cuantos;
            v >>= bits / 2;
            res += bits / 2;
         }
      }

      return (i == 0 && get_bit(v, 0) ? res : std::numeric_limits<T>::digits);
   }


   template<typename T>
   inline unsigned pop_first_set_bit(T& v)
   {
      auto bit = find_first_set_bit(v);
      reset_bit(v, bit);

      return bit;
   }

   template<typename T>
   inline unsigned pop_last_set_bit(T& v)
   {
      auto bit = find_last_set_bit(v);
      reset_bit(v, bit);

      return bit;
   }

   template<typename T>
   inline unsigned pop_ith_set_bit(T& v, unsigned i)
   {
      auto bit = find_ith_set_bit(v);
      reset_bit(v, bit);

      return bit;
   }


   namespace impl {
      template<typename T>
      struct referencia {
      public:
         referencia(T& m, std::size_t i)
         : memoria_(m), indice_(i)
         {
         }

         void operator=(bool b)
         {
            write_bit(memoria_, indice_, b);
         }

         void operator&=(bool b)
         {
            and_bit(memoria_, indice_, b);
         }

         void operator|=(bool b)
         {
            or_bit(memoria_, indice_, b);
         }

         void operator^=(bool b)
         {
            xor_bit(memoria_, indice_, b);
         }

         operator bool( ) const
         {
            return get( );
         }

         bool get( ) const
         {
            return get_bit(memoria_, indice_);
         }

         void reset( )
         {
            reset_bit(memoria_, indice_);
         }

         void set( )
         {
            set_bit(memoria_, indice_);
         }

         void flip( )
         {
            flip_bit(memoria_, indice_);
         }

         T& block( )
         {
            return memoria_;
         }

         const T& block( ) const
         {
            return memoria_;
         }

         const std::size_t& offset( ) const
         {
            return indice_;
         }

      private:
         T& memoria_;
         std::size_t indice_;
      };

      template<typename T>
      struct referencia_const {
      public:
         referencia_const(const T& m, std::size_t i)
         : memoria_(m), indice_(i)
         {
         }

         operator bool( ) const
         {
            return get( );
         }

         bool get( ) const
         {
            return get_bit(memoria_, indice_);
         }

         const T& block( ) const
         {
            return memoria_;
         }

         const std::size_t& offset( ) const
         {
            return indice_;
         }

      private:
         const T& memoria_;
         std::size_t indice_;
      };

      template<typename B>
      class bitset_base : private B {
      public:
         using block_type = typename B::block_type;
         using reference = referencia<block_type>;
         using const_reference = referencia_const<block_type>;

         template<typename... PARAMS>
         bitset_base(PARAMS&&... p)
         : B(std::forward<PARAMS>(p)...)
         {
            reset( );
         }

         static constexpr std::size_t bits_per_block( )
         {
            return B::bits_per_block( );
         }

         void set( )
         {
            std::fill(block_begin( ), block_end( ), ~block_type(0));

            if (B::trailing_bits( ) != 0) {
               reset_n_bits(*std::prev(block_end( )), bits_per_block( ) - B::trailing_bits( ), B::trailing_bits( ));
            }
         }

         void set_n(std::size_t i, std::size_t n)
         {
            aplica_n_(i, n, FUNCTOR(set_n_bits));
         }

         void reset( )
         {
            std::fill(block_begin( ), block_end( ), block_type(0));
         }

         void reset_n(std::size_t i, std::size_t n)
         {
            aplica_n_(i, n, FUNCTOR(reset_n_bits));
         }

         void flip( )
         {
            std::transform(block_begin( ), block_end( ), block_begin( ), std::bit_not<>( ));

            if (B::trailing_bits( ) != 0) {
               reset_n_bits(*std::prev(block_end( )), bits_per_block( ) - B::trailing_bits( ), B::trailing_bits( ));
            }
         }

         void flip_n(std::size_t i, std::size_t n)
         {
            aplica_n_(i, n, FUNCTOR(flip_n_bits));
         }

         std::size_t count( ) const
         {
            std::size_t res = 0;

            for (const auto& actual : make_range(block_begin( ), block_end( ))) {
               res += popcount(actual);
            }

            return res;
         }

         bool any( ) const
         {
            return std::any_of(block_begin( ), block_end( ), std::bind(std::not_equal_to<>( ), std::placeholders::_1, block_type(0)));
         }

         bool all( ) const
         {
            auto hay_padding = (B::trailing_bits( ) != 0);

            if (!std::all_of(block_begin( ), block_end( ) - hay_padding, std::bind(std::equal_to<>( ), std::placeholders::_1, ~block_type(0)))) {
               return false;
            }

            return !hay_padding || popcount(*std::prev(block_end( ))) == bits_per_block( ) - B::trailing_bits( );
         }

         bool none( ) const
         {
            return !any( );
         }

         block_type* block_begin( )
         {
            return B::block_begin( );
         }

         block_type* block_end( )
         {
            return B::block_end( );
         }

         std::reverse_iterator<block_type*> block_rbegin( )
         {
            return B::block_rbegin( );
         }

         std::reverse_iterator<block_type*> block_rend( )
         {
            return B::block_rend( );
         }

         const block_type* block_begin( ) const
         {
            return B::block_begin( );
         }

         const block_type* block_end( ) const
         {
            return B::block_end( );
         }

         std::reverse_iterator<block_type*> block_rbegin( ) const
         {
            return B::block_rbegin( );
         }

         std::reverse_iterator<block_type*> block_rend( ) const
         {
            return B::block_rend( );
         }

         reference operator[](std::size_t i)
         {
            return reference(block_begin( )[i / bits_per_block( )], i % bits_per_block( ));
         }

         const_reference operator[](std::size_t i) const
         {
            return const_reference(block_begin( )[i / bits_per_block( )], i % bits_per_block( ));
         }

         void operator&=(const bitset_base& b)
         {
            std::transform(block_begin( ), block_end( ), b.block_begin( ), block_begin( ), std::bit_and<>( ));
         }

         void operator|=(const bitset_base& b)
         {
            std::transform(block_begin( ), block_end( ), b.block_begin( ), block_begin( ), std::bit_or<>( ));
         }

         void operator^=(const bitset_base& b)
         {
            std::transform(block_begin( ), block_end( ), b.block_begin( ), block_begin( ), std::bit_xor<>( ));
         }

         std::size_t find_first_set( ) const
         {
            return primero_desde_(block_begin( ));
         }

         std::size_t find_last_set( ) const
         {
            return ultimo_hasta_(block_end( ));
         }

         std::size_t find_next_set(std::size_t i) const
         {
            auto ai = block_begin( ) + i / bits_per_block( );
            auto rel = find_next_set_bit(*ai, i % bits_per_block( ));

            return (rel != bits_per_block( ) ? (ai - block_begin( )) * bits_per_block( ) + rel : primero_desde_(ai + 1));
         }

         std::size_t find_previous_set(std::size_t i) const
         {
            auto ai = block_begin( ) + i / bits_per_block( );
            auto rel = find_previous_set_bit(*ai, i % bits_per_block( ));

            return (rel != bits_per_block( ) ? (ai - block_begin( )) * bits_per_block( ) + rel : ultimo_hasta_(ai));
         }

         std::size_t find_ith_set(std::size_t i) const
         {
            std::size_t res = 0;

            for (const auto& actual : make_range(block_begin( ), block_end( ))) {
               auto cuantos = popcount(actual);

               if (cuantos > i) {
                  return res + find_ith_set_bit(actual, i);
               }

               i -= cuantos;
               res += bits_per_block( );
            }

            return B::size( );
         }

         template<typename F>
         void visit(const F& f) const
         {
            for (std::size_t i = 0; i < B::blocks( ); ++i) {
               auto actual = block_begin( )[i];

               while (actual != 0) {
                  f(i * bits_per_block( ) + pop_first_set_bit(actual));
               }
            }
         }

      private:
         std::size_t primero_desde_(const block_type* ini) const
         {
            auto iter = std::find_if(ini, block_end( ), std::bind1st(std::not_equal_to<>( ), block_type(0)));
            return (iter != block_end( ) ? (iter - block_begin( )) * bits_per_block( ) + find_first_set_bit(*iter) : B::size( ));
         }

         std::size_t ultimo_hasta_(const block_type* fin) const
         {
            auto iter = find_last_if(block_begin( ), fin, std::bind1st(std::not_equal_to<>( ), block_type(0)));
            return (iter != block_end( ) ? (iter - block_begin( )) * bits_per_block( ) + find_last_set_bit(*iter) : B::size( ));
         }

         template<typename F>
         void aplica_n_(std::size_t i, std::size_t n, const F& f)
         {
            if (n == 0) {
               return;
            }

            auto ai = block_begin( ) + i / bits_per_block( );
            auto ax = i % bits_per_block( );
            auto m = std::min(n, bits_per_block( ) - ax);

            f(*ai++, ax, m);
            n -= m;

            auto completos = n / bits_per_block( );

            for (std::size_t i = 0; i < completos; ++i) {
               f(*ai++, 0, bits_per_block( ));
            }

            n -= completos * bits_per_block( );

            if (n != 0) {
               f(*ai, 0, n);
            }
         }
      };

      template<typename B>
      inline bitset_base<B> operator&(const bitset_base<B>& a, const bitset_base<B>& b)
      {
         auto temp = a;
         temp &= b;

         return temp;
      }

      template<typename B>
      inline bitset_base<B> operator|(const bitset_base<B>& a, const bitset_base<B>& b)
      {
         auto temp = a;
         temp |= b;

         return temp;
      }

      template<typename B>
      inline bitset_base<B> operator^(const bitset_base<B>& a, const bitset_base<B>& b)
      {
         auto temp = a;
         temp ^= b;

         return temp;
      }

      template<typename B>
      inline bitset_base<B> operator~(const bitset_base<B>& a)
      {
         auto temp = a;
         temp.flip( );

         return temp;
      }

      template<typename B>
      inline bool operator==(const bitset_base<B>& a, const bitset_base<B>& b)
      {
         return std::equal(a.block_begin( ), a.block_end( ), b.block_begin( ));
      }

      template<typename B>
      inline bool operator!=(const bitset_base<B>& a, const bitset_base<B>& b)
      {
         return !(a == b);
      }


      template<std::size_t N, typename T = unsigned>
      class static_bitset {
         static_assert(std::is_unsigned<T>( ), "base::impl::dynamic_bitset");

      public:
         using block_type = T;

         static constexpr std::size_t bits_per_block( )
         {
            return std::numeric_limits<T>::digits;
         }

         static constexpr std::size_t size( )
         {
            return N;
         }

         static constexpr std::size_t trailing_bits( )
         {
            return multiple_underflow(N, bits_per_block( ));
         }

         static constexpr std::size_t blocks( )
         {
            return ceil_division(N, bits_per_block( ));
         }

         T* block_begin( )
         {
            return memoria_.begin( );
         }

         T* block_end( )
         {
            return memoria_.end( );
         }

         std::reverse_iterator<T*> block_rbegin( )
         {
            return memoria_.rbegin( );
         }

         std::reverse_iterator<T*> block_rend( )
         {
            return memoria_.rend( );
         }

         const T* block_begin( ) const
         {
            return memoria_.begin( );
         }

         const T* block_end( ) const
         {
            return memoria_.end( );
         }

         std::reverse_iterator<const T*> block_rbegin( ) const
         {
            return memoria_.rbegin( );
         }

         std::reverse_iterator<const T*> block_rend( ) const
         {
            return memoria_.rend( );
         }

      private:
         std::array<T, blocks( )> memoria_;
      };

      template<typename T = unsigned, typename ALLOC = std::allocator<T>>
      class dynamic_bitset {
         static_assert(std::is_unsigned<T>( ), "base::impl::dynamic_bitset");

      public:
         using block_type = T;
         using allocator_type = ALLOC;

         dynamic_bitset(std::uint32_t n)
         : dynamic_bitset(n, ALLOC( ))
         {
         }

         dynamic_bitset(std::uint32_t n, ALLOC&& a)
         : tam_(n), memoria_(make_unique_array<T>(a, blocks( )))
         {
         }

         static constexpr std::size_t bits_per_block( )
         {
            return std::numeric_limits<T>::digits;
         }

         std::size_t size( ) const
         {
            return tam_;
         }

         std::size_t trailing_bits( ) const
         {
            return multiple_underflow(tam_, bits_per_block( ));
         }

         std::size_t blocks( ) const
         {
            return ceil_division(tam_, bits_per_block( ));
         }

         T* block_begin( )
         {
            return memoria_.get( );
         }

         T* block_end( )
         {
            return memoria_.get( ) + blocks( );
         }

         std::reverse_iterator<T*> block_rbegin( )
         {
            return std::make_reverse_iterator(memoria_.get( ) + blocks( ));
         }

         std::reverse_iterator<T*> block_rend( )
         {
            return std::make_reverse_iterator(memoria_.get( ));
         }

         const T* block_begin( ) const
         {
            return memoria_.get( );
         }

         const T* block_end( ) const
         {
            return memoria_.get( ) + blocks( );
         }

         std::reverse_iterator<const T*> block_rbegin( ) const
         {
            return std::make_reverse_iterator(memoria_.get( ) + blocks( ));
         }

         std::reverse_iterator<const T*> block_rend( ) const
         {
            return std::make_reverse_iterator(memoria_.get( ));
         }

      private:
         std::size_t tam_;
         unique_ptr<T[], ALLOC> memoria_;
      };

      template<typename T = unsigned>
      class adapted_bitset {
         static_assert(std::is_unsigned<T>( ), "base::impl::adapted_bitset");

      public:
         using block_type = T;

         adapted_bitset(std::uint32_t n, block_type* m)
         : tam_(n), memoria_(m)
         {
         }

         static constexpr std::size_t bits_per_block( )
         {
            return std::numeric_limits<T>::digits;
         }

         std::size_t size( ) const
         {
            return tam_;
         }

         std::size_t trailing_bits( ) const
         {
            return multiple_underflow(tam_, bits_per_block( ));
         }

         std::size_t blocks( ) const
         {
            return ceil_division(tam_, bits_per_block( ));
         }

         T* block_begin( )
         {
            return memoria_;
         }

         T* block_end( )
         {
            return memoria_ + blocks( );
         }

         std::reverse_iterator<T*> block_rbegin( )
         {
            return std::make_reverse_iterator(memoria_ + blocks( ));
         }

         std::reverse_iterator<T*> block_rend( )
         {
            return std::make_reverse_iterator(memoria_);
         }

         const T* block_begin( ) const
         {
            return memoria_;
         }

         const T* block_end( ) const
         {
            return memoria_ + blocks( );
         }

         std::reverse_iterator<const T*> block_rbegin( ) const
         {
            return std::make_reverse_iterator(memoria_ + blocks( ));
         }

         std::reverse_iterator<const T*> block_rend( ) const
         {
            return std::make_reverse_iterator(memoria_);
         }

      private:
         std::size_t tam_;
         block_type* memoria_;
      };
   }

   template<std::size_t N, typename T = unsigned>
   using bitset = impl::bitset_base<impl::static_bitset<N, T>>;

   template<typename T = unsigned, typename ALLOC = std::allocator<T>>
   using dynamic_bitset = impl::bitset_base<impl::dynamic_bitset<T, ALLOC>>;

   template<typename T = unsigned>
   using adapted_bitset = impl::bitset_base<impl::adapted_bitset<T>>;
}

#endif
