#ifndef BASE_FUNCTIONAL_H
#define BASE_FUNCTIONAL_H

#include "intrinsic.h"
#include <limits>
#include <type_traits>
#include <utility>

namespace base {
   template<typename F>
   class reverse_order {
   public:
      constexpr reverse_order( )
      {
      }

      constexpr reverse_order(F f)
      : func_(std::move(f))
      {
      }

      template<typename T1, typename T2>
      constexpr bool operator()(T1&& p1, T2&& p2) const
      {
         return func_(std::forward<T2>(p2), std::forward<T1>(p1));
      }

   private:
      F func_;
   };

   template<typename F>
   inline reverse_order<F> make_reverse_order(F f)
   {
      return { std::move(f) };
   }


   namespace impl {
      template<typename PRED>
      class not_fun {
      public:
         explicit not_fun(PRED&& p)
         : pred_(std::move(p))
         {
         }

         explicit not_fun(const PRED& p)
         : pred_(p)
         {
         }

         template<typename... PARAMS>
         bool operator()(PARAMS&&... p)
         {
            return !pred_(std::forward<PARAMS>(p)...);
         }

      private:
         PRED pred_;
      };
   }

   template<typename PRED>
   auto not_fn(PRED&& p)
   {
      return impl::not_fun<std::decay_t<PRED>>(std::forward<PRED>(p));
   }


   template<typename T>
   constexpr auto empty(const T& c)
   {
      return c.empty( );
   }

   template<typename T, std::size_t N>
   constexpr auto empty(const T(&arr)[N])
   {
      return N == 0;
   }

   template<typename T>
   constexpr auto empty(std::initializer_list<T> i)
   {
      return i.size( ) == 0;
   }


   namespace impl {
      template<typename T, std::size_t... I, typename... PARAMS>
      inline void invoca(T& t, std::index_sequence<I...> i, PARAMS&&... p)
      {
         using expand_type = int[];
         expand_type{ (std::get<I>(t)(p...), 0)... };
      }
   }

   template<typename... F>
   class function_set {
   public:
      function_set(F&&... f)
      : t_(std::forward<F>(f)...)
      {
      }

      template<typename... PARAMS>
      void operator()(PARAMS&&... p) const
      {
         impl::invoca(t_, std::make_index_sequence<sizeof...(F)>( ), std::forward<PARAMS>(p)...);
      }

   private:
      std::tuple<F...> t_;
   };

   template<typename... F>
   inline auto make_function_set(F&&... f)
   {
      return function_set<F...>{ std::forward<F>(f)... };
   }
}

#endif
