#ifndef BASE_INTEGER_H
#define BASE_INTEGER_H

#include <algorithm>
#include <climits>
#include <cstdint>
#include <limits>
#include <type_traits>

namespace base {
	template<typename T1, typename T2>
   inline constexpr typename std::common_type<T1, T2>::type multiple_overflow(const T1& n, const T2& d)
   {
      return n % d;
   }

	template<typename T1, typename T2>
   inline constexpr typename std::common_type<T1, T2>::type multiple_underflow(const T1& n, const T2& d)
   {
      return (d - (n % d)) % d;
   }


	template<typename T1, typename T2>
   inline constexpr typename std::common_type<T1, T2>::type floor_division(const T1& n, const T2& d)
   {
      return n / d;
   }

	template<typename T1, typename T2>
   inline constexpr typename std::common_type<T1, T2>::type ceil_division(const T1& n, const T2& d)
   {
      return n / d + bool(n % d);
   }


   template<typename T>
   inline constexpr T log2(const T& n)
   {
      return (n == 1 ? 0 : 1 + log2(T(n / 2)));
   }

   inline constexpr unsigned log2(unsigned char n)
   {
      return std::numeric_limits<unsigned>::digits - __builtin_clz(n) - 1;
   }

   inline constexpr unsigned log2(unsigned short n)
   {
      return std::numeric_limits<unsigned>::digits - __builtin_clz(n) - 1;
   }

   inline constexpr unsigned log2(unsigned n)
   {
      return std::numeric_limits<unsigned>::digits - __builtin_clz(n) - 1;
   }

   inline constexpr unsigned log2(unsigned long n)
   {
      return std::numeric_limits<unsigned long>::digits - __builtin_clzl(n) - 1;
   }

   inline constexpr unsigned log2(unsigned long long n)
   {
      return std::numeric_limits<unsigned long long>::digits - __builtin_clzll(n) - 1;
   }

   #ifdef __SIZEOF_INT128__
      inline constexpr unsigned log2(unsigned __int128 n)
      {
         auto m2 = std::uint64_t(n >> 64);

         if (m2 != 0) {
            return 64 + log2(m2);
         }

         auto m1 = std::uint64_t(n);
         return log2(m1);
      }
   #endif

   namespace impl {
      template<typename T>
      inline constexpr T pow2(const T& n, std::false_type f)
      {
         return (n == 0 ? 1 : 2 * pow2(T(n - 1), f));
      }

      template<typename T>
      inline constexpr T pow2(const T& n, std::true_type t)
      {
         return T(1) << n;
      }

      template<typename T>
      inline constexpr bool is_power2(const T& n, std::false_type f)
      {
         return n != 0 && n == pow2(log2(n), f);
      }

      template<typename T>
      inline constexpr bool is_power2(const T& n, std::true_type t)
      {
         return n != 0 && !(n & (n - 1));
      }
   }

   template<typename T>
   inline constexpr T pow2(const T& n)
   {
      return impl::pow2(n, std::is_integral<T>( ));
   }

   template<typename T>
   inline constexpr bool is_power2(const T& n)
   {
      return impl::is_power2(n, std::is_integral<T>( ));
   }

   template<typename T>
   inline constexpr T next_power2(const T& n)
   {
      return pow2(T(log2(n) + 1));
   }

   template<typename T>
   inline constexpr T ceil_power2(const T& n)
   {
      return (is_power2(n) ? n : next_power2(n));
   }

   template<typename T>
   inline constexpr T bits_for_semirange(const T& n)
   {
      return (n != 0 ? log2(n) + !is_power2(n) : 0);
   }

   template<typename T>
   inline constexpr T bits_for_count(const T& n)
   {
      return bits_for_semirange(n + 1);
   }

   template<typename T>
   inline constexpr T bytes_for_semirange(const T& n)
   {
      return base::ceil_division(bits_for_semirange(n), 8);
   }

   template<typename T>
   inline constexpr T bytes_for_count(const T& n)
   {
      return bytes_for_semirange(n + 1);
   }


   template<std::size_t B>
   struct integer;

   template<>
   struct integer<8> {
      using type = std::int8_t;
   };

   template<>
   struct integer<16> {
      using type = std::int16_t;
   };

   template<>
   struct integer<32> {
      using type = std::int32_t;
   };

   template<>
   struct integer<64> {
      using type = std::int64_t;
   };

   #ifdef __SIZEOF_INT128__
      template<>
      struct integer<128> {
         using type = __int128;
      };
   #endif


   template<std::size_t B>
   struct unsigned_integer;

   template<>
   struct unsigned_integer<8> {
      using type = std::uint8_t;
   };

   template<>
   struct unsigned_integer<16> {
      using type = std::uint16_t;
   };

   template<>
   struct unsigned_integer<32> {
      using type = std::uint32_t;
   };

   template<>
   struct unsigned_integer<64> {
      using type = std::uint64_t;
   };

   #ifdef __SIZEOF_INT128__
      template<>
      struct unsigned_integer<128> {
         using type = unsigned __int128;
      };
   #endif


   template<std::size_t B>
   using int_t = typename integer<B>::type;

   template<std::size_t B>
   using uint_t = typename unsigned_integer<B>::type;


   template<std::size_t B>
   using int_least_t = typename integer<std::max(ceil_power2(std::size_t(B)), std::size_t(8))>::type;

   template<std::size_t B>
   using uint_least_t = typename unsigned_integer<std::max(ceil_power2(std::size_t(B)), std::size_t(8))>::type;


   template<std::size_t N>
   using int_count_t = int_least_t<bits_for_count(std::size_t(N)) + 1>;

   template<std::size_t N>
   using uint_count_t = uint_least_t<bits_for_count(std::size_t(N))>;


   template<std::size_t N>
   using int_semirange_t = int_least_t<bits_for_semirange(std::size_t(N)) + 1>;

   template<std::size_t N>
   using uint_semirange_t = uint_least_t<bits_for_semirange(std::size_t(N))>;


   #ifdef __SIZEOF_INT128__
      using intmax_t = int_t<std::max(std::size_t(128), CHAR_BIT * sizeof(std::uintmax_t))>;
      using uintmax_t = uint_t<std::max(std::size_t(128), CHAR_BIT * sizeof(std::uintmax_t))>;
   #else
      using intmax_t = std::intmax_t;
      using uintmax_t = std::uintmax_t;
   #endif
}

#endif
