#ifndef BASE_INTRINSIC_H
#define BASE_INTRINSIC_H

#include <iterator>
#include <memory>
#include <type_traits>
#include <utility>

namespace base {
   template<typename T>
   inline void create(T& v)
   {
      new(&v) T;
   }

   template<typename T, typename... PARAMS>
   inline void create(T& v, PARAMS&&... p)
   {
      new(&v) T(std::forward<PARAMS>(p)...);
   }

   template<typename T>
   inline void destroy(T& v)
   {
      v.~T( );
   }


   template<typename T>
   inline void raw_move(T& a, T& b)
   {
      destroy(a);
      create(a, std::move(b));
   }

   template<typename T>
   inline void move(T& a, T& b)
   {
      if (&a != &b) {
         raw_move(a, b);
      }
   }


   template<typename T>
   inline void raw_assign(T& a, const T& b)
   {
      T temp(b);

      destroy(a);
      create(a, std::move(temp));
   }

   template<typename T>
   inline void assign(T& a, const T& b)
   {
      if (&a != &b) {
         raw_assign(a, b);
      }
   }


   template<typename T, typename ALLOC>
   inline T* dynamic_allocate_hint(ALLOC a, std::size_t n, const void* h)
   {
      typename ALLOC::template rebind<T>::other delegado(a);
      return delegado.allocate(n, h);
   }

   template<typename T, typename ALLOC>
   inline T* dynamic_allocate(ALLOC a, std::size_t n)
   {
      return dynamic_allocate_hint<T>(a, n, nullptr);
   }

   template<typename T, typename ALLOC>
   inline T* dynamic_create_hint(ALLOC a, const void* h)
   {
      auto res = dynamic_allocate_hint<T>(a, 1, h);
      create(*res);

      return res;
   }

   template<typename T, typename ALLOC>
   inline T* dynamic_create(ALLOC a)
   {
      return dynamic_create_hint<T>(a, nullptr);
   }

   template<typename T, typename ALLOC>
   inline T* dynamic_create_array_hint(ALLOC a, std::size_t n, const void* h)
   {
      auto res = dynamic_allocate_hint<T>(a, n, h);

      for (std::size_t i = 0; i < n; ++i) {
         create(res[i]);
      }

      return res;
   }

   template<typename T, typename ALLOC>
   inline T* dynamic_create_array(ALLOC a, std::size_t n)
   {
      return dynamic_create_array_hint<T>(a, n, nullptr);
   }


   template<typename T, typename ALLOC>
   inline void dynamic_deallocate(ALLOC a, T* p, std::size_t n)
   {
      typename ALLOC::template rebind<T>::other delegado(a);
      delegado.deallocate(p, n);
   }

   template<typename T, typename ALLOC>
   inline void dynamic_destroy(ALLOC a, T* p)
   {
      destroy(*p);
      dynamic_deallocate(a, p, 1);
   }

   template<typename T, typename ALLOC>
   inline void dynamic_destroy_array(ALLOC a, T* p, std::size_t n)
   {
      for (std::size_t i = n; i > 0; --i) {
         destroy(p[i - 1]);
      }

      dynamic_deallocate(a, p, n);
   }


   namespace impl {
      template<typename A>
      class allocator_deleter : private A {
      public:
         allocator_deleter( ) = default;
         allocator_deleter(A a)
         : A(std::move(a))
         {
         }

         template<typename T>
         void operator()(T* p)
         {
            dynamic_destroy(static_cast<A&>(*this), p);
         }
      };

      template<typename A, bool T>
      class array_allocator_deleter;

      template<typename A>
      class array_allocator_deleter<A, false> : private A {
      public:
         array_allocator_deleter( ) = default;
         array_allocator_deleter(A a, std::size_t n)
         : A(std::move(a)), n_(n)
         {
         }

         template<typename T>
         void operator()(T* p)
         {
            dynamic_destroy_array(static_cast<A&>(*this), p, n_);
         }

      private:
         std::size_t n_;
      };

      template<typename A>
      class array_allocator_deleter<A, true> : private A {
      public:
         array_allocator_deleter( ) = default;
         array_allocator_deleter(A a, std::size_t n)
         : A(std::move(a))
         {
         }

         template<typename T>
         void operator()(T* p)
         {
            dynamic_destroy_array(static_cast<A&>(*this), p, 0);
         }
      };
   }

   template<typename T, typename ALLOC>
   using unique_ptr = typename std::conditional<!std::is_array<T>::value,
      std::unique_ptr<T, impl::allocator_deleter<ALLOC>>,
      std::unique_ptr<T, impl::array_allocator_deleter<ALLOC, std::is_trivially_destructible<typename std::remove_extent<T>::type>::value>>
   >::type;

   template<typename T, typename ALLOC>
   inline unique_ptr<T, ALLOC> make_unique(ALLOC a)
   {
      return { dynamic_create<T>(a), a };
   }

   template<typename T, typename ALLOC>
   inline unique_ptr<T[], ALLOC> make_unique_array(ALLOC a, std::size_t n)
   {
      return { dynamic_create_array<T>(a, n), { a, n } };
   }


	template<typename ITER>
	class range {
   public:
		range(ITER i, ITER f)
		: i_(i), f_(f)
		{
		}

		bool empty( ) const
		{
         return i_ == f_;
		}

		std::size_t size( ) const
		{
         return std::distance(i_, f_);
		}

		ITER begin( ) const
		{
			return i_;
		}

		ITER end( ) const
		{
			return f_;
		}

		auto& operator[](std::size_t i) const
		{
         return begin( )[i];
		}

	private:
		ITER i_;
		ITER f_;
	};

	template<typename ITER>
	inline range<ITER> make_range(ITER i, ITER f)
	{
		return { i, f };
	}


   #define FUNCTOR(f)                                              \
      [](auto&&... v) { return f(std::forward<decltype(v)>(v)...); }
}

#endif
