#ifndef BASE_MEMORY_H
#define BASE_MEMORY_H

#include "bits.h"
#include "integer.h"
#include "intrinsic.h"
#include <cstddef>
#include <iterator>
#include <type_traits>

namespace base {
   template<typename T>
   class underlying_type {
   public:
      using value_type = T;

      value_type& get( )
      {
         return reinterpret_cast<T&>(mem_);
      }

      const value_type& get( ) const
      {
         return reinterpret_cast<const T&>(mem_);
      }

   private:
      typename std::aligned_storage<sizeof(T), alignof(T)>::type mem_;
   };


   inline constexpr std::uintptr_t disguise(std::uintptr_t p)
   {
      return p;
   }

   inline constexpr std::uintptr_t disguise(std::uintptr_t p, std::uintptr_t bandera)
   {
      return p | bandera;
   }

   inline constexpr std::uintptr_t disguise(std::nullptr_t p)
   {
      return reinterpret_cast<std::uintptr_t>(p);
   }

   template<typename T>
   inline constexpr std::uintptr_t disguise(T* p)
   {
      return reinterpret_cast<std::uintptr_t>(p);
   }

   template<typename T>
   inline constexpr std::uintptr_t disguise(T* p, std::uintptr_t bandera)
   {
      return disguise(reinterpret_cast<std::uintptr_t>(p), bandera);
   }

   template<typename T>
   inline constexpr std::uintptr_t disguise_flag(std::uintptr_t p)
   {
      static_assert(std::is_pointer<T>( ), "base::disguise_flag");
      return bitmask_until(p, log2(alignof(typename std::remove_pointer<T>::type)));
   }

   template<typename T>
   inline constexpr T undisguise(std::uintptr_t p)
   {
      static_assert(std::is_pointer<T>( ), "base::undisguise");
      return reinterpret_cast<T>(bitmask_from(p, log2(alignof(typename std::remove_pointer<T>::type))));
   }


   template<typename T>
   inline unsigned char* get_raw_byte_iterator(T& v)
   {
      return reinterpret_cast<unsigned char*>(&v);
   }

   template<typename T>
   inline const unsigned char* get_raw_byte_iterator(const T& v)
   {
      return reinterpret_cast<const unsigned char*>(&v);
   }

   template<typename T>
   inline std::reverse_iterator<unsigned char*> get_reverse_raw_byte_iterator(T& v)
   {
      return std::reverse_iterator<unsigned char*>(reinterpret_cast<unsigned char*>(&v) + sizeof(T));
   }

   template<typename T>
   inline std::reverse_iterator<const unsigned char*> get_reverse_raw_byte_iterator(const T& v)
   {
      return std::reverse_iterator<const unsigned char*>(reinterpret_cast<const unsigned char*>(&v) + sizeof(T));
   }

   using raw_byte_iterator = decltype(get_raw_byte_iterator(std::declval<int&>( )));
   using const_raw_byte_iterator = decltype(get_raw_byte_iterator(std::declval<const int&>( )));

   using reverse_raw_byte_iterator = decltype(get_reverse_raw_byte_iterator(std::declval<int&>( )));
   using const_reverse_raw_byte_iterator = decltype(get_reverse_raw_byte_iterator(std::declval<const int&>( )));

   #ifdef __BYTE_ORDER__
      inline constexpr bool is_big_endian( )
      {
         return
            #if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
               true
            #elif __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
               false
            #endif
         ;
      }

      inline constexpr bool is_little_endian( )
      {
         return
            #if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
               false
            #elif __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
               true
            #endif
         ;
      }

      template<typename T>
      inline auto get_most_significant_byte_iterator(T&& v)
      {
         return
            #if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
               get_raw_byte_iterator(std::forward<T>(v))
            #elif __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
               get_reverse_raw_byte_iterator(std::forward<T>(v))
            #endif
         ;
      }

      template<typename T>
      inline auto get_least_significant_byte_iterator(T&& v)
      {
         return
            #if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
               get_reverse_raw_byte_iterator(std::forward<T>(v))
            #elif __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
               get_raw_byte_iterator(std::forward<T>(v))
            #endif
         ;
      }

      using most_significant_byte_iterator = decltype(get_most_significant_byte_iterator(std::declval<int&>( )));
      using const_most_significant_byte_iterator = decltype(get_most_significant_byte_iterator(std::declval<const int&>( )));

      using least_significant_byte_iterator = decltype(get_least_significant_byte_iterator(std::declval<int&>( )));
      using const_least_significant_byte_iterator = decltype(get_least_significant_byte_iterator(std::declval<const int&>( )));
   #endif
}

#endif
