#ifndef GREVLEX_H
#define GREVLEX_H

#include "monomial.h"
#include "base/array.h"
#include "base/bits.h"
#include <tbb/parallel_for.h>
#include <algorithm>
#include <cstdint>
#include <vector>

struct grevlex {
   template<std::uint32_t N>
   bool operator()(const monomial<N>& a, const monomial<N>& b) const
   {
      auto ac = a.degree( );
      auto bc = b.degree( );

      if (ac != bc) {
         return ac < bc;
      }

      for (std::uint32_t i = ac; i > 0; --i) {
         auto ai = a[i - 1];
         auto bi = b[i - 1];

         if (ai != bi) {
            return ai > bi;
         }
      }

      return false;
   }
};

template<std::uint32_t N>
class grevlex_converter {
public:
   grevlex_converter( )
   {
      for (std::uint32_t f = 0; f <= N; ++f) {
         for (std::uint32_t u = 0; u <= N; ++u) {
            cuenta_[f][u] = (u > f ? 0 : (u == 0 || u == f ? 1 : cuenta_[f - 1][u] + cuenta_[f - 1][u - 1]));
         }
      }

      for (std::uint32_t i = 0; i <= N + 1; ++i) {
         acumulado_previo_[i] = (i == 0 ? 0 : acumulado_previo_[i - 1] + cuenta_[N][i - 1]);
      }
   }

   std::uint32_t degree_of(std::uint32_t i) const
   {
      return std::upper_bound(acumulado_previo_.begin( ), acumulado_previo_.end( ), i) - acumulado_previo_.begin( ) - 1;
   }

   std::uint32_t degree_of(const monomial<N>& m) const
   {
      return m.degree( );
   }

   monomial<N> monomial_of(std::uint32_t i) const
   {
      auto grado = degree_of(i);
      return genera_monomio_(inverso_(i - population_before(grado), grado), grado);
   }

   std::uint32_t index_of(const monomial<N>& m) const
   {
      auto grado = degree_of(m);
      return inverso_(genera_indice_(m, grado), grado) + population_before(grado);
   }

   std::uint32_t population_before(std::uint32_t d) const
   {
      return acumulado_previo_[d];
   }

   std::uint32_t population_up_to(std::uint32_t d) const
   {
      return acumulado_previo_[d + 1];
   }

   std::uint32_t population_of(std::uint32_t d) const
   {
      return population_up_to(d) - population_before(d);
   }

private:
   std::uint32_t inverso_(std::uint32_t i, std::uint32_t d) const
   {
      return population_of(d) - i - 1;
   }

   monomial<N> genera_monomio_(std::uint32_t i, std::uint32_t u) const
   {
      auto m = monomial<N>( );
      auto f = std::uint32_t(N);

      while (i != 0) {
         auto temp = cuenta_[--f][u];

         if (i >= temp) {
            m.insert(f);

            i -= temp;
            u -= 1;
         }
      }

      for (std::uint32_t i = 0; i < u; ++i) {
         m.insert(i);
      }

      return m;
   }

   std::uint32_t genera_indice_(const monomial<N>& m, std::uint32_t d) const
   {
      auto i = std::uint32_t(0);
      auto u = std::uint32_t(0);

      m.visit([&](auto pos) {
         i += cuenta_[pos][++u];
      });

      return i;
   }

   base::matrix<std::uint32_t, N + 1, N + 1> cuenta_;
   std::array<std::uint32_t, N + 2> acumulado_previo_;
};

template<std::uint32_t N>
class grevlex_cache : public grevlex_converter<N> {
public:
   grevlex_cache( )
   {
   }

   grevlex_cache(std::uint32_t d)
   {
      emplace_degree(d);
   }

   auto degree_of(const monomial<N>& m) const
   {
      return static_cast<const grevlex_converter<N>*>(this)->degree_of(m);
   }

   auto degree_of(std::uint32_t i) const
   {
      return monomios_[i].degree( );
   }

   auto monomial_of(std::uint32_t i) const
   {
      return monomios_[i];
   }

   void emplace_degree(std::uint32_t d)
   {
      auto actuales = monomios_.size( );
      monomios_.resize(grevlex_converter<N>::population_up_to(d));

      tbb::parallel_for(actuales, monomios_.size( ), [this](auto i) {
         monomios_[i] = static_cast<const grevlex_converter<N>*>(this)->monomial_of(i);
      });
   }

private:
   std::vector<monomial<N>> monomios_;
};

#endif
