#ifndef MONOMIAL_H
#define MONOMIAL_H

#include "base/algorithm.h"
#include "base/bits.h"
#include "base/functional.h"
#include <iterator>

template<std::uint32_t N>
class monomial {
public:
   monomial( )
   {
   }

   static constexpr std::uint32_t capacity( )
   {
      return N;
   }

   std::uint32_t degree( ) const
   {
      return bits_.count( );
   }

   std::uint32_t operator[](std::uint32_t i) const
   {
      return bits_.find_ith_set(i);
   }

   bool none( ) const
   {
      return bits_.none( );
   }

   bool find(std::uint32_t i) const
   {
      return bits_[i];
   }

   void insert(std::uint32_t i)
   {
      bits_[i].set( );
   }

   void erase(std::uint32_t i)
   {
      bits_[i].reset( );
   }

   void clear( )
   {
      bits_.reset( );
   }

   template<typename F>
   void visit(const F& f) const
   {
      bits_.visit(f);
   }

   monomial& operator&=(const monomial& m)
   {
      bits_ &= m.bits_;
      return *this;
   }

   monomial& operator*=(const monomial& m)
   {
      bits_ |= m.bits_;
      return *this;
   }

   monomial& operator/=(const monomial& m)
   {
      bits_ ^= m.bits_;
      return *this;
   }

   monomial operator~( ) const
   {
      auto res = *this;
      res.bits_ = ~res.bits_;

      return res;
   }

   bool operator==(const monomial& m) const
   {
      return bits_ == m.bits_;
   }

   bool operator!=(const monomial& m) const
   {
      return bits_ != m.bits_;
   }

private:
   base::bitset<N> bits_;
};

template<std::uint32_t N>
inline monomial<N> operator&(monomial<N>&& a, const monomial<N>& b)
{
   a &= b;
   return std::move(a);
}

template<std::uint32_t N>
inline monomial<N> operator&(const monomial<N>& a, const monomial<N>& b)
{
   return monomial<N>(a) & b;
}

template<std::uint32_t N>
inline monomial<N> operator*(monomial<N>&& a, const monomial<N>& b)
{
   a *= b;
   return std::move(a);
}

template<std::uint32_t N>
inline monomial<N> operator*(const monomial<N>& a, const monomial<N>& b)
{
   return monomial<N>(a) * b;
}

template<std::uint32_t N>
inline monomial<N> operator/(monomial<N>&& a, const monomial<N>& b)
{
   a /= b;
   return std::move(a);
}

template<std::uint32_t N>
inline monomial<N> operator/(const monomial<N>& a, const monomial<N>& b)
{
   return monomial<N>(a) / b;
}

template<std::uint32_t N>
inline monomial<N> least_common_multiple(const monomial<N>& a, const monomial<N>& b)
{
   return a * b;
}

template<std::uint32_t N>
inline monomial<N> greatest_common_divisor(const monomial<N>& a, const monomial<N>& b)
{
   return a & b;
}

template<std::uint32_t N>
inline bool is_divisible(const monomial<N>& a, const monomial<N>& b)
{
   return least_common_multiple(a, b) == a;
}

template<std::uint32_t N>
inline bool is_coprime(const monomial<N>& a, const monomial<N>& b)
{
   return greatest_common_divisor(a, b).none( );
}

namespace impl {
   template<std::uint32_t N, typename RI, typename F, typename OP>
   inline void set_visit(const monomial<N>& m, RI ai, RI af, const F& f, const OP& op)
   {
      for (auto i = ai; i != af; ++i) {
         auto actual = op(m, *i);

         if (f(actual)) {
            set_visit(actual, i + 1, af, f, op);
         }
      }
   }

   template<std::uint32_t N, typename RI, typename F, typename OP>
   inline bool set_find(const monomial<N>& m, RI ai, RI af, const F& f, const OP& op)
   {
      for (auto i = ai; i != af; ++i) {
         auto actual = op(m, *i);
         auto res = f(actual);

         if (res.first || res.second && set_find(actual, i + 1, af, f, op)) {
            return true;
         }
      }

      return false;
   }

   template<std::uint32_t N, typename F, typename V, typename OP>
   inline auto set_walk(const monomial<N>& m, const monomial<N>& bandera, const F& f, const V& v, const OP& op)
   {
      std::array<std::uint32_t, N> prendidos;
      auto escribir = prendidos.begin( );

      bandera.visit([&](auto pos) {
         *escribir++ = pos;
      });

      return v(m, prendidos.begin( ), escribir, f, op);
   }

   template<std::uint32_t N, typename F, typename V>
   inline auto subset_walk(const monomial<N>& m, const F& f, const V& v)
   {
      return set_walk(m, m, f, v, [](auto m, auto i) {
         return m.erase(i), m;
      });
   }

   template<std::uint32_t N, typename F, typename V>
   inline auto superset_walk(const monomial<N>& m, const F& f, const V& v)
   {
      return set_walk(m, ~m, f, v, [](auto m, auto i) {
         return m.insert(i), m;
      });
   }
}

template<std::uint32_t N, typename F>
inline void subset_visit(const monomial<N>& m, const F& f)
{
   impl::subset_walk(m, f, FUNCTOR(impl::set_visit));
}

template<std::uint32_t N, typename F>
inline bool subset_find(const monomial<N>& m, const F& f)
{
   return impl::subset_walk(m, f, FUNCTOR(impl::set_find));
}

template<std::uint32_t N, typename F>
inline void superset_visit(const monomial<N>& m, const F& f)
{
   impl::superset_walk(m, f, FUNCTOR(impl::set_visit));
}

template<std::uint32_t N, typename F>
inline bool superset_find(const monomial<N>& m, const F& f)
{
   return impl::superset_walk(m, f, FUNCTOR(impl::set_find));
}

#endif
