#ifndef POLYNOMIAL_H
#define POLYNOMIAL_H

#include "simd.h"
#include "base/algorithm.h"
#include "base/bits.h"
#include "base/functional.h"
#include "base/integer.h"
#include <tbb/scalable_allocator.h>
#include <algorithm>
#include <climits>
#include <cstdint>
#include <functional>

template<typename T>
inline bool empty_segment(const T& v)
{
   return is_zero(v);
}

template<typename T>
inline void clear_segment(T& destino)
{
   destino = T{ };
}

template<typename T>
inline void exclusive_merge_segment(T& destino, const T& origen)
{
   destino ^= origen;
}

class polynomial {
public:
   polynomial( ) = default;

   explicit polynomial(std::uint32_t capacidad)
   : mem_(base::make_unique_array<vector_type>(tbb::scalable_allocator<vector_type>( ), vectors(capacidad)))
   {
      clear(capacidad);
   }

   static constexpr std::uint32_t bits_per_block( )
   {
      return sizeof(block_type) * CHAR_BIT;
   }

   static constexpr std::uint32_t bits_per_vector( )
   {
      return sizeof(vector_type) * CHAR_BIT;
   }

   static std::uint32_t blocks(std::uint32_t capacidad)
   {
      return base::ceil_division(capacidad, bits_per_block( ));
   }

   static std::uint32_t vectors(std::uint32_t capacidad)
   {
      return base::ceil_division(capacidad, bits_per_vector( ));
   }

   bool empty( ) const
   {
      return mem_.get( ) == nullptr;
   }

   void reset( )
   {
      mem_.reset( );
   }

   void clear(std::uint32_t capacidad)
   {
      std::for_each(vector_begin( ), vector_end(capacidad), FUNCTOR(clear_segment));
   }

   bool find(std::uint32_t i) const
   {
      return base::get_bit(block_begin( )[i / bits_per_block( )], i % bits_per_block( ));
   }

   void insert(std::uint32_t i)
   {
      base::set_bit(block_begin( )[i / bits_per_block( )], i % bits_per_block( ));
   }

   void erase(std::uint32_t i)
   {
      base::reset_bit(block_begin( )[i / bits_per_block( )], i % bits_per_block( ));
   }

   void insert_erase(std::uint32_t i)
   {
      base::flip_bit(block_begin( )[i / bits_per_block( )], i % bits_per_block( ));
   }

   void exclusive_merge(std::uint32_t capacidad, const polynomial& p)
   {
      std::transform(vector_begin( ), vector_end(capacidad), p.vector_begin( ), vector_begin( ), std::bit_xor<>( ));
   }

   std::uint32_t find_leader(std::uint32_t capacidad) const
   {
      auto ai = vector_begin( );
      auto af = vector_end(capacidad);
      auto aj = base::find_last_if(ai, af, base::not_fn(FUNCTOR(empty_segment)));

      if (aj == af) {
         return -1;
      }

      auto bi = reinterpret_cast<const block_type*>(aj);

      for (std::uint32_t i = bits_per_vector( ) / bits_per_block( ); i > 0; --i) {
         auto local = bi[i - 1];

         if (local != 0) {
            return bits_per_vector( ) * std::uint32_t(aj - ai) + bits_per_block( ) * std::uint32_t(i - 1) + base::find_last_set_bit(local);
         }
      }
   }

   template<typename F>
   void visit(std::uint32_t capacidad, const F& f) const
   {
      auto ai = vector_begin( );
      auto af = vector_end(capacidad);
      auto aj = ai;

      while (aj = std::find_if(aj, af, base::not_fn(FUNCTOR(empty_segment))), aj != af) {
         auto bi = reinterpret_cast<const block_type*>(aj);

         for (std::uint32_t i = 0; i < bits_per_vector( ) / bits_per_block( ); ++i) {
            auto local = bi[i];

            while (local != 0) {
               f(bits_per_vector( ) * std::uint32_t(aj - ai) + bits_per_block( ) * i + base::pop_first_set_bit(local));
            }
         }

         ++aj;
      }
   }

   block_type* block_begin( )
   {
      return reinterpret_cast<block_type*>(mem_.get( ));
   }

   block_type* block_end(std::uint32_t capacidad)
   {
      return reinterpret_cast<block_type*>(mem_.get( )) + blocks(capacidad);
   }

   const block_type* block_begin( ) const
   {
      return reinterpret_cast<const block_type*>(mem_.get( ));
   }

   const block_type* block_end(std::uint32_t capacidad) const
   {
      return reinterpret_cast<const block_type*>(mem_.get( )) + blocks(capacidad);
   }

   vector_type* vector_begin( )
   {
      return mem_.get( );
   }

   vector_type* vector_end(std::uint32_t capacidad)
   {
      return mem_.get( ) + vectors(capacidad);
   }

   const vector_type* vector_begin( ) const
   {
      return mem_.get( );
   }

   const vector_type* vector_end(std::uint32_t capacidad) const
   {
      return mem_.get( ) + vectors(capacidad);
   }

private:
   base::unique_ptr<vector_type[], tbb::scalable_allocator<vector_type>> mem_;
};

#endif
