#ifndef SPARSE_POLYNOMIAL_H
#define SPARSE_POLYNOMIAL_H

#include "simd.h"
#include "base/bits.h"
#include "base/integer.h"
#include "base/memory.h"
#include <tbb/scalable_allocator.h>
#include <algorithm>
#include <climits>
#include <cstdint>
#include <iterator>

namespace impl {
   inline constexpr std::pair<std::uint32_t, std::uint32_t> memoria_completo(std::uint8_t b)
   {
      std::pair<std::uint32_t, std::uint32_t> res(0, 0);

      for (auto i = 1; i <= b; ++i) {
         res.first *= 256;
         res.first += 1;

         res.second *= 256;
         res.second += 32;
      }

      return res;
   }

   inline constexpr std::pair<std::uint32_t, std::uint32_t> memoria_maxima(std::uint32_t v)
   {
      auto c4 = base::get_n_bits(v, 24, 8);
      auto c3 = base::get_n_bits(v, 16, 8);
      auto c2 = base::get_n_bits(v,  8, 8);
      auto c1 = base::get_n_bits(v,  0, 8);

      constexpr auto m3 = memoria_completo(3);
      constexpr auto m2 = memoria_completo(2);
      constexpr auto m1 = memoria_completo(1);
      constexpr auto m0 = memoria_completo(0);

      std::pair<std::uint32_t, std::uint32_t> res(0, 0);

      res.first += c4 * m3.first + 1;
      res.first += c3 * m2.first + 1;
      res.first += c2 * m1.first + 1;
      res.first += c1 * m0.first + 1;

      res.second += c4 * m3.second + std::min(c4 + 1, std::uint32_t(32));
      res.second += c3 * m2.second + std::min(c3 + 1, std::uint32_t(32));
      res.second += c2 * m1.second + std::min(c2 + 1, std::uint32_t(32));
      res.second += c1 * m0.second + std::min(c1 + 1, std::uint32_t(32));

      return { res.first, res.second + base::multiple_underflow(res.second, 32) };
   }


   static_assert(base::is_little_endian( ), "sparse_polynomial");

   inline void escribe_bits(std::pair<std::uint8_t*, std::uint8_t>& datos1, std::uint8_t v)
   {
      *reinterpret_cast<std::uint16_t*>(datos1.first) |= std::uint16_t(v) << datos1.second;

      datos1.second += 5;
      datos1.first += (datos1.second >= 8);
      datos1.second %= 8;
   }

   inline auto lee_bits(std::pair<const std::uint8_t*, std::uint8_t>& datos1)
   {
      auto res = base::get_n_bits(*reinterpret_cast<const std::uint16_t*>(datos1.first), datos1.second, 5);

      datos1.second += 5;
      datos1.first += (datos1.second >= 8);
      datos1.second %= 8;

      return res;
   }


   inline std::uint8_t* comprime_headers(std::uint8_t* ai, std::uint8_t* af)
   {
      std::pair<std::uint8_t*, std::uint8_t> res(ai, 0);

      for (; ai != af; ++ai) {
         auto actual = *ai;
         base::reset_bits(*ai);

         escribe_bits(res, actual);
      }

      return res.first + bool(res.second);
   }

   inline bool procesa_bitset(std::reverse_iterator<std::uint8_t*>& datos1, std::reverse_iterator<std::uint8_t*>& datos2, std::reverse_iterator<std::uint8_t*>& datos3, const base::bitset<256, block_type>& v)
   {
      auto cuantos = std::uint16_t(v.count( ));

      if (cuantos != 0) {
         *datos1++ = cuantos * (cuantos < 32);

         if (cuantos < 32) {
            datos2 += cuantos;
            auto copia = datos2.base( );

            v.visit([&](auto i) {
               *copia++ = i;
            });
         }
         else {
            datos3 += 32;
            auto& bitset = *reinterpret_cast<base::bitset<256, block_type>*>(datos3.base( ));

            bitset = v;
         }
      }

      return cuantos;
   }

   inline std::uint8_t* comprime(std::reverse_iterator<std::uint8_t*>& datos1, std::reverse_iterator<std::uint8_t*>& datos2, std::reverse_iterator<std::uint8_t*>& datos3, const block_type* pi, const block_type* pf)
   {
      constexpr auto bits_per_block = sizeof(block_type) * CHAR_BIT;

      auto datos1_ini = datos1;
      auto offset = std::uint32_t(pf - pi) * bits_per_block;

      base::bitset<256, block_type> v1;
      base::bitset<256, block_type> v2;
      base::bitset<256, block_type> v3;
      base::bitset<256, block_type> v4;

      while (offset != 0) {
         offset -= bits_per_block;
         v1.block_begin( )[(offset % 256) / bits_per_block] = *--pf;

         if (offset % 256 == 0) {
            if (procesa_bitset(datos1, datos2, datos3, v1)) {
               v1.reset( );
               v2[base::get_n_bits(offset, 8, 8)].set( );
            }

            if (offset % 65536 == 0) {
               if (procesa_bitset(datos1, datos2, datos3, v2)) {
                  v2.reset( );
                  v3[base::get_n_bits(offset, 16, 8)].set( );
               }

               if (offset % 16777216 == 0) {
                  if (procesa_bitset(datos1, datos2, datos3, v3)) {
                     v3.reset( );
                     v4[base::get_n_bits(offset, 24, 8)].set( );
                  }
               }
            }
         }
      }

      procesa_bitset(datos1, datos2, datos3, v4);
      return comprime_headers(datos1.base( ), datos1_ini.base( ));
   }


   template<std::uint8_t B>
   struct descomprime_impl {
      template<typename F>
      void operator()(std::pair<const std::uint8_t*, std::uint8_t>& datos1, const std::uint8_t*& datos2, const std::uint8_t*& datos3, const F& f, std::uint32_t prefijo)
      {
         auto cuantos = impl::lee_bits(datos1);

         if (cuantos != 0) {
            auto copia = datos2;
            datos2 += cuantos;

            for (auto i = 0; i < cuantos; ++i) {
               descomprime_impl<B - 1>( )(datos1, datos2, datos3, f, prefijo | std::uint32_t(copia[i]) << (8 * (B - 1)));
            }
         }
         else {
            auto& bitset = *reinterpret_cast<const base::bitset<256, block_type>*>(datos3);
            datos3 += 32;

            bitset.visit([&](auto i) {
               descomprime_impl<B - 1>( )(datos1, datos2, datos3, f, prefijo | std::uint32_t(i) << (8 * (B - 1)));
            });
         }
      }
   };

   template<>
   struct descomprime_impl<1> {
      template<typename F>
      void operator()(std::pair<const std::uint8_t*, std::uint8_t>& datos1, const std::uint8_t*& datos2, const std::uint8_t*& datos3, const F& f, std::uint32_t prefijo)
      {
         auto cuantos = impl::lee_bits(datos1);

         if (cuantos != 0) {
            auto copia = datos2;
            datos2 += cuantos;

            f(prefijo, copia, datos2);
         }
         else {
            auto& bitset = *reinterpret_cast<const base::bitset<256, block_type>*>(datos3);
            datos3 += 32;

            f(prefijo, bitset);
         }
      }
   };

   template<typename F>
   inline void descomprime(const std::uint8_t* d1, const std::uint8_t* datos2, const std::uint8_t* datos3, const F& f)
   {
      std::pair<const std::uint8_t*, std::uint8_t> datos1(d1, 0);
      descomprime_impl<4>( )(datos1, datos2, datos3, f, 0);
   }


   template<typename F>
   struct visitante {
      void operator()(std::uint32_t prefijo, const base::bitset<256, block_type>& b) const
      {
         b.visit([&, this](auto i) {
            f(prefijo | i);
         });
      }

      void operator()(std::uint32_t prefijo, const std::uint8_t* ini, const std::uint8_t* fin) const
      {
         while (ini != fin) {
            f(prefijo | *ini++);
         }
      }

      const F& f;
   };
}

class sparse_assign {
public:
	sparse_assign(polynomial& p, std::uint32_t c)
	: p_(p), capacidad_(c)
	{
	}

   void operator()(std::uint32_t valor) const
   {
      p_.insert_erase(valor);
   }

   void operator()(std::uint32_t prefijo, const std::uint8_t* ini, const std::uint8_t* fin) const
   {
      while (ini != fin) {
         p_.insert(prefijo | *ini++);
      }
   }

   void operator()(std::uint32_t prefijo, const base::bitset<256, block_type>& b) const
   {
      auto pi = p_.block_begin( ) + prefijo / p_.bits_per_block( );
      auto pf = p_.block_begin( ) + prefijo / p_.bits_per_block( ) + 256 / p_.bits_per_block( );

      std::copy_n(b.block_begin( ), std::min(pf, p_.block_end(capacidad_)) - pi, pi);
   }

private:
   polynomial& p_;
   std::uint32_t capacidad_;
};

class sparse_exclusive_merge {
public:
	sparse_exclusive_merge(polynomial& p, std::uint32_t c)
	: p_(p), capacidad_(c)
	{
	}

   void operator()(std::uint32_t valor) const
   {
      p_.insert_erase(valor);
   }

   void operator()(std::uint32_t prefijo, const std::uint8_t* ini, const std::uint8_t* fin) const
   {
      while (ini != fin) {
         p_.insert_erase(prefijo | *ini++);
      }
   }

   void operator()(std::uint32_t prefijo, const base::bitset<256, block_type>& b) const
   {
      auto pi = p_.block_begin( ) + prefijo / p_.bits_per_block( );
      auto pf = p_.block_begin( ) + prefijo / p_.bits_per_block( ) + 256 / p_.bits_per_block( );

      std::transform(pi, std::min(pf, p_.block_end(capacidad_)), b.block_begin( ), pi, std::bit_xor<>( ));
   }

private:
   polynomial& p_;
   std::uint32_t capacidad_;
};

class sparse_polynomial {
public:
   sparse_polynomial( ) = default;
   sparse_polynomial(const polynomial& p, std::uint32_t capacidad)
   {
      auto mem = impl::memoria_maxima(capacidad);
      alignas(block_type) std::uint8_t memoria_h[mem.first];
      alignas(block_type) std::uint8_t memoria_c[mem.second];
      alignas(block_type) std::uint8_t memoria_v[mem.second];

      auto datos1 = std::make_reverse_iterator(memoria_h + mem.first);
      auto datos2 = std::make_reverse_iterator(memoria_c + mem.second);
      auto datos3 = std::make_reverse_iterator(memoria_v + mem.second);
      auto fin_headers = impl::comprime(datos1, datos2, datos3, p.block_begin( ), p.block_end(capacidad));

      auto tam_header = datos1 - std::make_reverse_iterator(fin_headers);
      auto tam_octeto = datos2 - std::make_reverse_iterator(memoria_c + mem.second);
      auto tam_vector = datos3 - std::make_reverse_iterator(memoria_v + mem.second);
      mem_ = base::make_unique_array<block_type>(tbb::scalable_allocator<block_type>( ), base::ceil_division(offset_( ) + tam_header + tam_octeto + tam_vector, sizeof(block_type)));
      guarda_datos_(tam_header, tam_octeto, tam_vector);

      auto info = buffer_hcv_( );
      std::copy(datos1.base( ), fin_headers, std::get<0>(info));
      std::copy(datos2.base( ), memoria_c + mem.second, std::get<1>(info));
      std::copy(datos3.base( ), memoria_v + mem.second, std::get<2>(info));
   }

   bool empty( ) const
   {
      return mem_.get( ) == nullptr;
   }

   void reset( )
   {
      mem_.reset( );
   }

   template<typename F>
   void visit(const F& f) const
   {
      auto info = buffer_hcv_( );
      impl::descomprime(std::get<0>(info), std::get<1>(info), std::get<2>(info), impl::visitante<F>{f});
   }

   template<typename F>
   void leaf_visit(const F& f) const
   {
      auto info = buffer_hcv_( );
      impl::descomprime(std::get<0>(info), std::get<1>(info), std::get<2>(info), f);
   }

private:
   static constexpr std::uint32_t offset_( )
   {
      return std::max(2 * sizeof(std::uint32_t), sizeof(block_type));
   }

   void guarda_datos_(std::uint32_t tam_header, std::uint32_t tam_octeto, std::uint32_t tam_vector)
   {
      reinterpret_cast<std::uint32_t*>(mem_.get( ))[0] = tam_vector;
      reinterpret_cast<std::uint32_t*>(mem_.get( ))[1] = tam_octeto;
   }

   std::tuple<std::uint8_t*, std::uint8_t*, std::uint8_t*> buffer_hcv_( ) const
   {
      auto tam_vector = reinterpret_cast<const std::uint32_t*>(mem_.get( ))[0];
      auto tam_octeto = reinterpret_cast<const std::uint32_t*>(mem_.get( ))[1];
      auto buffer = reinterpret_cast<std::uint8_t*>(mem_.get( )) + offset_( );

      return std::make_tuple(buffer + tam_vector + tam_octeto, buffer + tam_vector, buffer);
   }

   base::unique_ptr<block_type[], tbb::scalable_allocator<block_type>> mem_;
};

#endif
